"""# Hierarchical Markov Clustering (HMC) with Gamma bus detection."""
from textwrap import dedent
from typing import List, Optional, Tuple, Union

from ragraph.analysis import _utils
from ragraph.analysis._classes import ClusterAnalysis, Parameter
from ragraph.analysis.bus._gamma import gamma_bus_detection, gamma_params
from ragraph.analysis.cluster._markov import (
    hierarchical_markov,
    markov_params,
)
from ragraph.graph import Graph, Node

markov_gamma_params = {
    p.name: p
    for p in [
        Parameter(
            "local_buses",
            bool,
            "Whether to detect buses locally (instead of globally).",
            default=True,
        )
    ]
}
markov_gamma_params.update(**gamma_params, **markov_params)

markov_gamma_analysis = ClusterAnalysis(
    "Hierarchical Markov clustering with Gamma bus detection",
    description=dedent(
        """\
    Cluster a given graph hierarchically with bus detection on a local level or
    globally.

    Note:
        1. A hierarchy is initialized using Hierarchical Markov Clustering.
        2. For all top-level clusters, a bus is detected.
            a. If a bus is found, both the bus and non-bus are re-clustered.
        3. Move down a level and go back to 2.
            b. If at bottom-level, stop.
    """
    ),
    parameters=markov_gamma_params,
)


@markov_gamma_analysis
def markov_gamma(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    leafs: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    alpha: int = markov_gamma_params["alpha"].default,  # type: ignore
    beta: float = markov_gamma_params["beta"].default,  # type: ignore
    mu: float = markov_gamma_params["mu"].default,  # type: ignore
    gamma: float = markov_gamma_params["gamma"].default,  # type: ignore
    local_buses: bool = markov_gamma_params["local_buses"].default,  # type: ignore
    max_iter: int = markov_gamma_params["max_iter"].default,  # type: ignore
    symmetrize: bool = markov_gamma_params["symmetrize"].default,  # type: ignore
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Union[List[Node], Tuple[Graph, List[Node]]]:
    """Cluster a given graph hierarchically with buses on a local level or globally."""
    assert leafs is not None
    graph, roots = hierarchical_markov(
        graph,
        root=root,
        leafs=leafs,
        inherit=inherit,
        loops=loops,
        alpha=alpha,
        beta=beta,
        mu=mu,
        edge_weights=edge_weights,
        inplace=True,
        max_iter=max_iter,
        symmetrize=symmetrize,
        names=False,
        safe=False,
    )
    check_roots = roots[:]
    checked_roots = []
    while check_roots:
        local_root = check_roots.pop()
        local_leafs = [leaf for leaf in leafs if leaf in local_root.descendants]

        # Detect bus in current root.
        bus_leafs, nonbus_leafs = gamma_bus_detection(
            graph,
            local_root,
            leafs=local_leafs,
            inherit=inherit,
            loops=loops,
            gamma=gamma,
            edge_weights=edge_weights,
            names=False,
            safe=False,
        )

        # Bus detected!
        if bus_leafs:
            # Unset local hierarchy.
            _utils.clear_local_hierarchy(graph, local_leafs, [local_root])
            local_root.children = None

            # Recalculate the nonbus hierarchy.
            graph, nonbus_roots = hierarchical_markov(
                graph,
                alpha=alpha,
                beta=beta,
                mu=mu,
                root=None,
                leafs=nonbus_leafs,
                inherit=inherit,
                loops=loops,
                edge_weights=edge_weights,
                inplace=True,
                max_iter=max_iter,
                symmetrize=symmetrize,
                names=False,
                safe=False,
            )
            _utils.set_children(graph, local_root, nonbus_roots)

            # Recalculate the bus hierarchy.
            graph, bus_roots = hierarchical_markov(
                graph,
                alpha=alpha,
                beta=beta,
                mu=mu,
                root=None,
                leafs=bus_leafs,
                inherit=inherit,
                loops=loops,
                edge_weights=edge_weights,
                inplace=True,
                max_iter=max_iter,
                symmetrize=symmetrize,
                names=False,
                safe=False,
            )
            local_root.children = bus_roots + local_root.children
            for bus in bus_roots:
                bus.is_bus = True

        # Finished this root.
        checked_roots.append(local_root)

        # If local: move down a level.
        if not check_roots and local_buses:
            check_roots = [child for local_root in checked_roots for child in local_root.children]
            checked_roots = []

    return graph, roots



"""Sort nodes in a typical axis-ready format."""
from textwrap import dedent
from typing import Any, Dict, Iterable, List, Optional, Tuple, Union

from ragraph.analysis._classes import Cast, Parameter, SequenceAnalysis
from ragraph.graph import Graph
from ragraph.node import Node


class KindsCast(Cast):
    def __init__(self, **kwargs):
        pass

    def __call__(self, value: Any) -> Optional[List[str]]:
        if value is None:
            return None
        elif not isinstance(value, list):
            raise ValueError(f"Node kind order should be a list, got '{value}'.")
        else:
            return value


axis_sequencing_analysis = SequenceAnalysis(
    "Plot axis sequencing. Get nodes in a plot-axis-ready order.",
    description=dedent(
        """Order nodes for a typical display on matrix plot axis. They are sorted by:
          1. Node kinds in order of (supplied) occurrence.
          2. Top-down hierarchy.
          3. Optional: Bus-nodes before nonbus-nodes in w.r.t. siblings.
          4. Optional: Node width (e.g. size in terms of leaf nodes).

        Note:
            Axis ready means ordered by kind and buses nodes first where possible.

        Note:
            `inherit`, `loops`, and `edge_weights` are ignored for this analysis.

        Warning:
            When supplying node kinds it's your own responsibility to make sure that each node's
            kind actually is included in that node kind order.
        """
    ),
    parameters={
        "kinds": Parameter(
            "kinds",
            Optional[List[str]],
            description="Optional order of node kinds.",
            cast=KindsCast(),
            default=None,
        ),
        "sort_by_bus": Parameter(
            "sort_by_bus",
            bool,
            description="Whether to put bus nodes first.",
            cast=bool,
            default=True,
        ),
        "sort_by_width": Parameter(
            "sort_by_width",
            bool,
            description="Whether to sort nodes by width in terms of leaf nodes.",
            cast=bool,
            default=True,
        ),
    },
)


def get_axis_sequence(
    nodes: Iterable[Node],
    kinds: Optional[List[str]] = None,
    sort_by_bus: bool = True,
    sort_by_width: bool = True,
    root_cache: Optional[Dict[Node, Node]] = None,
    child_cache: Optional[Dict[Node, List[Node]]] = None,
    width_cache: Optional[Dict[Node, int]] = None,
) -> List[Node]:
    """Get nodes in a plot-axis-ready order.

    Arguments:
        nodes: Axis nodes to order.
        kinds: Optional order of node kinds.
        sort_by_bus: Whether to put bus nodes first.
        sort_by_width: Whether to sort nodes by width in terms of leaf nodes.
        root_cache: Node to resulting root node cache. Any supplied dictionary
            will be updated in-place.
        child_cache: Node to child nodes that have been seen cache.
        width_cache: Node to node width cache.

    Returns:
        Ordered nodes.

    Note:
        Axis ready means ordered by kind and buses nodes first where possible.
        Any supplied cache dictionary will be updated in-place.

    Warning:
        When supplying node kinds it's your own responsibility to make sure
        that each node's kind actually is included in that node kind order.
    """
    root_cache = dict() if root_cache is None else root_cache
    child_cache = dict() if child_cache is None else child_cache
    width_cache = dict() if width_cache is None else width_cache

    kinds = get_kinds(nodes) if kinds is None else kinds
    kind_order = {kind: position for position, kind in enumerate(kinds)}

    roots = get_roots(nodes, root_cache, child_cache)
    roots.sort(key=lambda x: get_root_key(x, kind_order, sort_by_width, width_cache))

    # Sort by is_bus status with buses first, width second, if applicable.
    if sort_by_bus or sort_by_width:
        for parent in child_cache.keys():
            child_cache[parent].sort(
                key=lambda x: get_sibling_key(x, sort_by_bus, sort_by_width, width_cache)
            )

    # Complement width_cache for roots for completeness sake.
    if sort_by_width:
        for root in roots:
            width_cache[root] = sum(
                width_cache.get(child, 1) for child in child_cache.get(root, [])
            )

    leafs = [leaf for root in roots for leaf in get_leafs(root, child_cache)]
    return leafs


def get_root(
    node: Node,
    root_cache: Optional[Dict[Node, Node]] = None,
    child_cache: Optional[Dict[Node, List[Node]]] = None,
) -> Node:
    """Get root node of a node.

    Arguments:
        node: Nodes to find corresponding root nodes for.
        root_cache: Node to resulting root node cache. Any supplied dictionary
            will be updated in-place.
        child_cache: Node to child nodes that have been seen cache. Any supplied
            dictionary will be updated in-place.

    Returns:
        Root node.

    Note:
        Supply cache dictionaries to store intermediate results.
    """

    # Cache from any seen node to the resulting root node.
    root_cache = dict() if root_cache is None else root_cache

    # Cache from any parent we came across to the included children in the search.
    child_cache = dict() if child_cache is None else child_cache

    # Cached, return cache value.
    if node in root_cache:
        return root_cache[node]

    # Is a root, return.
    if node.parent is None:
        return node

    # Find and update caches accordingly.
    if node.parent not in child_cache:
        child_cache[node.parent] = [node]
    else:
        child_cache[node.parent].append(node)
    root = get_root(node.parent, root_cache, child_cache)
    root_cache[node] = root

    return root


def get_roots(
    nodes: Iterable[Node],
    root_cache: Optional[Dict[Node, Node]] = None,
    child_cache: Optional[Dict[Node, List[Node]]] = None,
) -> List[Node]:
    """Get list of root nodes corresponding to some collection of nodes.

    Arguments:
        nodes: Collection of nodes to find corresponding root nodes for.
        root_cache: Node to resulting root node cache. Any supplied dictionary
            will be updated in-place.
        child_cache: Node to child nodes that have been seen cache. Any supplied
            dictionary will be updated in-place.

    Returns:
        Root nodes.

    Note:
        Supply cache dictionaries to store intermediate results.
    """
    # Cache from any seen node to the resulting root node.
    root_cache = dict() if root_cache is None else root_cache

    # Cache from any parent we came across to the included children in the search.
    child_cache = dict() if child_cache is None else child_cache

    seen = set()  # Seen root nodes.
    ordered = []  # Seen root nodes in order of appearance.
    for node in nodes:
        root = get_root(node, root_cache, child_cache)
        if root not in seen:
            seen.add(root)
            ordered.append(root)
    return ordered


def get_kinds(nodes: Iterable[Node]) -> List[str]:
    """Get node kinds in order of occurrence.

    Arguments:
        nodes: Nodes to iterate over.

    Returns:
        List of Node kinds.
    """
    ordered = []
    seen = set()
    for node in nodes:
        kind = node.kind
        if kind not in seen:
            ordered.append(kind)
            seen.add(kind)
    return ordered


def get_leafs(parent: Node, child_cache: Dict[Node, List[Node]]) -> List[Node]:
    """Get leaf nodes from a linked-list style child_cache (e.g. parent to children).

    Arguments:
        parent: Parent node.
        child_cache: Linked-list style parent-children dictionary.

    Returns:
        Displayed leaf nodes as they occur under this parent.
    """
    # Handle the case where this node is actually a leaf node.
    if parent not in child_cache:
        return [parent]

    # Fetch leaf nodes from deeper into the cached hierarchy.
    leafs = []
    for child in child_cache[parent]:
        leafs.extend(get_leafs(child, child_cache))
    return leafs


def get_width(node: Node, width_cache: Optional[Dict[Node, int]] = None) -> int:
    """Get width of a node and update width cache in-place.

    Arguments:
        node: Node to get the width of.
        width_cache: Node to node width cache.

    Returns:
        Node width.
    """
    width_cache = dict() if width_cache is None else width_cache
    if node not in width_cache:
        if node.children:
            width_cache[node] = sum(get_width(child, width_cache) for child in node.children)
        else:
            width_cache[node] = 1
    return width_cache[node]


def get_root_key(
    node: Node,
    kind_order: Dict[str, int],
    sort_by_width: bool,
    width_cache: Optional[Dict[Node, int]] = None,
) -> Union[int, Tuple[int, int]]:
    """Calculate a sorting score for root nodes.

    Arguments:
        node: Root node to obtain a score for.
        kind_order: Node kind to position dictionary.
        sort_by_width: Whether to sort nodes by width in terms of leaf nodes.
        width_cache: Node to node width cache.

    Returns:
        Root node sorting key.

    Note:
        Sort by kinds first, width in terms of actual (not display) leaf nodes second.
    """
    if sort_by_width:
        return (
            kind_order[node.kind],
            -get_width(node, width_cache),
        )
    else:
        return kind_order[node.kind]


def get_sibling_key(
    node: Node,
    sort_by_bus: bool,
    sort_by_width: bool,
    width_cache: Optional[Dict[Node, int]] = None,
) -> Union[int, Tuple[int, int]]:
    """Calculate a sorting score for sibling nodes. Lower scores go first.

    Arguments:
        node: Node to score.
        sort_by_bus: Whether to put bus nodes first.
        sort_by_width: Whether to sort nodes by width in terms of leaf nodes.
        width_cache: Dictionary of nodes to displayed children.

    Returns:
        Sibling node sorting score.

    Note:
        Bus nodes first, width in terms of actual (not display) leaf nodes second."""
    if sort_by_bus and sort_by_width:
        return (not node.is_bus, -get_width(node, width_cache))
    elif sort_by_bus:
        return not node.is_bus
    elif sort_by_width:
        return -get_width(node, width_cache)
    else:
        return 0


@axis_sequencing_analysis
def axis(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[Node], List[str]]] = None,
    kinds: Optional[List[str]] = None,
    sort_by_bus: bool = True,
    sort_by_width: bool = True,
    edge_weights: Optional[List[str]] = None,
    inherit: bool = True,
    loops: bool = False,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
):
    """docstring stub"""
    seq = get_axis_sequence(nodes, kinds, sort_by_bus, sort_by_width)  # type: ignore
    return graph, seq



"""# Legend plot component"""

from collections import defaultdict
from copy import deepcopy
from typing import List

import plotly.graph_objs as go

from ragraph.edge import Edge
from ragraph.plot.generic import Component, Style


class Legend(Component):
    """Legend component.

    Arguments:
        edges: List displayed edges.
        style: Plot style mapping.
    """

    def __init__(self, edges: List[Edge], style: Style = Style()):
        if style.piemap.display in ["kinds", "labels", "weight labels"]:
            # create categorical legend:
            categories = style.piemap.fields or get_legend_categories(
                edges=edges, display=style.piemap.display
            )

            max_label_length = max(len(cat) for cat in categories)
            fontsize = (
                int(0.6 * style.boxsize) if style.legend.fontsize is None else style.legend.fontsize
            )
            width = (
                max_label_length * fontsize * style.legend.fontaspectratio + 1 * style.boxsize
            )  # Text width plus 2 box sizes as margin.

            traces = get_categorical_legend_traces(categories=categories, style=style)
            xaxis, yaxis = deepcopy(style.legend.xaxis), deepcopy(style.legend.yaxis)
            num = len(categories)
            xaxis.update(
                range=[0, (width / style.boxsize) * style.xstep],
            )
            yaxis.update(
                range=[0, (num + 1) * style.ystep],
            )

            super().__init__(
                width=width,
                height=num * style.ystep * style.boxsize,
                traces=traces,
                shapes=None,
                xaxis=xaxis,
                yaxis=yaxis,
            )

        elif style.piemap.display == "weights":
            # Create numerical legend.
            traces = get_numerical_legend_traces(edges=edges, style=style)
            xaxis, yaxis = deepcopy(style.legend.xaxis), deepcopy(style.legend.yaxis)

            weight_labels = style.piemap.fields or get_legend_categories(
                edges=edges, display="weight labels"
            )

            max_label_length = max(len(label) for label in weight_labels)

            fontsize = (
                int(0.6 * style.boxsize) if style.legend.fontsize is None else style.legend.fontsize
            )

            width = 2.5 * style.xstep * len(weight_labels) * style.boxsize
            xaxis.update(
                range=[0, width / style.boxsize * style.xstep],
                scaleanchor="y",
                scaleratio=1.0,
            )
            yaxis.update(range=[0, (style.legend.height) * style.ystep])

            super().__init__(
                width=width,
                height=(style.legend.height) * style.boxsize,
                traces=traces,
                shapes=None,
                xaxis=xaxis,
                yaxis=yaxis,
            )


def get_categorical_legend_traces(
    categories: List[str], style: Style = Style()
) -> List[go.Scatter]:
    """Create categorical legend traces.

    Arguments:
        categories: List of categories.
        style: Plot style mapping.

    Returns:
        List of traces.
    """
    traces = []
    x0 = 0.5
    y0 = (len(categories) - 0.5) * style.ystep

    xdata = []
    ydata = []

    fontsize = int(0.6 * style.boxsize) if style.legend.fontsize is None else style.legend.fontsize

    for idx, cat in enumerate(categories):
        y = y0 - idx * style.ystep
        color = style.palettes.get_categorical_color(idx, field=cat)
        traces.append(
            go.Scatter(
                x=[x0],
                y=[y],
                text=[cat],
                mode="markers",
                marker=dict(color=color),
                hoverinfo="text",
                showlegend=False,
            )
        )

        xdata.append(x0 + 0.5 * style.xstep)
        ydata.append(y)

    traces.append(
        go.Scatter(
            x=xdata,
            y=ydata,
            text=categories,
            hoverinfo="text",
            mode="text",
            showlegend=False,
            textfont=dict(
                color=style.legend.fontcolor,
                family=style.legend.fontfamily,
                size=fontsize,
            ),
            textposition="middle right",
        )
    )

    return traces


def get_swatchtraces(
    idx: int,
    name: str,
    colors: List[str],
    min_value: float,
    max_value: float,
    style: Style = Style(),
) -> List[go.Scatter]:
    """Swatch bar of color map.

    Arguments:
        idx: Number of the swatch that must be created.
        name: name of the swatch trace.
        colors: List of colors to be added to the swatch trace
        min_value: Minimum value of the weight related to the color map.
        max_value: Maximmum value of the weight related to the color map.
        style: Style object.

    Returns:
        List of go.Scatter traces.
    """
    if style.legend.n_ticks < 2:
        x = (1.8 * idx + 0.75) * style.xstep
    else:
        x = (2.5 * idx + 0.75) * style.xstep

    ystep = style.legend.height * style.ystep / len(colors)
    traces = []

    for idc, color in enumerate(colors):
        x0 = x - 0.35 * style.xstep
        x1 = x + 0.35 * style.xstep
        xdata = [x0, x0, x1, x1]

        y0 = style.ystep + (idc) * ystep
        y1 = style.ystep + (idc + 1) * ystep
        ydata = [y0, y1, y1, y0]

        traces.append(
            go.Scatter(
                x=xdata,
                y=ydata,
                fill="toself",
                text=[name for x in xdata],
                mode="lines",
                hoverinfo="text",
                marker=dict(size=0),
                fillcolor=color,
                line=dict(color=color),
                showlegend=False,
            )
        )

    fontsize = int(0.6 * style.boxsize) if style.legend.fontsize is None else style.legend.fontsize

    if style.legend.n_ticks >= 2:
        n = style.legend.n_ticks
        tick_step = (max_value - min_value) / (n - 1)
        traces.append(
            go.Scatter(
                x=[x + 0.5 * style.xstep for step in range(n)],
                y=[style.ystep + step * style.legend.height / (n - 1) for step in range(n)],
                text=["{:.1f}".format(min_value + step * tick_step) for step in range(n)],
                showlegend=False,
                mode="text",
                textfont=dict(
                    color=style.legend.fontcolor,
                    family=style.legend.fontfamily,
                    size=0.7 * fontsize,
                ),
                hoverinfo="text",
                textposition="middle right",
            )
        )

    traces.append(
        go.Scatter(
            x=[x],
            y=[0.5 * style.ystep],
            mode="text",
            text=[name[0:3] + "."],
            showlegend=False,
            textfont=dict(
                color=style.legend.fontcolor,
                family=style.legend.fontfamily,
                size=fontsize,
            ),
            hoverinfo="text",
        )
    )
    return traces


def get_numerical_legend_traces(edges: List[Edge], style: Style = Style()) -> List[go.Scatter]:
    """Get traces for a numerical legend,

    Arguments:
        edges: List of edges that are displayed.
        style: Style object.

    Returns:
       List of traces.
    """
    weight_labels = style.piemap.fields or get_legend_categories(
        edges=edges, display="weight labels"
    )

    traces = []
    for idx, label in enumerate(weight_labels):
        colors = style.palettes.get_continuous_palette(label)
        edge_dict = defaultdict(list)
        for e in edges:
            if e.weights.get(label, None):
                edge_dict[e.source, e.target].append(e.weights[label])

        values = [sum(weights) for weights in edge_dict.values()]
        min_value = min(values)
        max_value = max(values)
        if min_value == max_value:
            min_value = max_value - 1

        traces.extend(
            get_swatchtraces(
                idx=idx,
                name=label,
                colors=colors,
                min_value=min_value,
                max_value=max_value,
                style=style,
            )
        )

    return traces


def get_legend_categories(edges: List[Edge], display: str) -> List[str]:
    """Get the number of categories to be included in the legend.

    Arguments:
        edges: The list of edges that are being displayed.
        display: The selected display mode.

    Returns:
        list of categories.
    """
    if display == "kinds":
        return sorted(list(set([e.kind for e in edges])))
    elif display == "labels":
        labels = set()
        for e in edges:
            labels.update(set(e.labels))
        return sorted(list(labels))
    elif display == "weight labels":
        weight_labels = set()
        for e in edges:
            weight_labels.update(set([w for w in e.weights]))
        return sorted(list(weight_labels))
    else:
        return []



"""# Compatibility analysis

Calculate compatibility between different variants for several functional elements.
"""

from copy import deepcopy
from itertools import combinations
from math import prod
from pathlib import Path
from typing import Callable, Dict, Generator, Iterable, List, Optional, Tuple, Union

from plotly import graph_objs as go

from ragraph.analysis.compatibility.bdd import yield_feasible_configurations
from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node
from ragraph.plot import generic, mdm

try:  # pragma: no cover
    import numpy as np

    HAVE_NUMPY = True
except ImportError:
    HAVE_NUMPY = False


def is_feasible(
    compat: Union["np.ndarray", List[List[Union[bool, int]]]], config: Tuple[int, ...]
) -> bool:
    """Return whether this configuration is feasible based on a compatibility matrix."""
    coords = combinations(config, 2)
    return all(compat[row][col] > 0 for row, col in coords)


def get_variant_score(
    variant: Node, weights: Optional[List[str]] = None, aggregation: str = "sum"
) -> float:
    """Score a single variant node as an aggregation of selected weights.

    Arguments:
        variant: Variant node.
        weights: Optional list of weights to take into account.
        aggregation: Aggregation method, either "sum" or "product".

    Returns:
        Variant score.
    """
    values = [variant.weights[k] for k in weights] if weights else variant.weights.values()
    if aggregation == "sum":
        return sum(values)
    return prod(values)


def get_configuration_score(variant_scores: Iterable[float], aggregation: str = "sum") -> float:
    """Score a configuration by aggregating variant scores by either taking the
    sum or product.

    Arguments:
        variant_scores: List of scores for all selected variants.

    Returns:
        Config score.
    """
    if aggregation == "sum":
        return sum(variant_scores)
    return prod(variant_scores)


def get_interface_method(
    interface_kind: Optional[str] = None,
) -> Callable[[Graph, Node, Node], bool]:
    """Get a basic interface checking method. Will check for the existence of edges
    between the ascendants of the variants (optionally filtered with an edge kind).
    """

    def has_interface(graph: Graph, var1: Node, var2: Node) -> bool:
        return any(
            e
            for e in graph.edges_between_all(var1.ancestor_gen, var2.ancestor_gen)
            if interface_kind is None or e.kind == interface_kind
        )

    return has_interface


def get_compatibility_method(
    compatibility_kind: Optional[str] = "compatibility",
    incompatibility_kind: Optional[str] = None,
) -> Callable[[Graph, Node, Node], bool]:
    """Get a method dat signals whether two variants are compatible.

    Arguments:
        compatibility_kind: Optional Edge kind that signals variant compatibility.
        incompatibility_kind: Optional Edge kind that signals variant incompatibility.

    Returns:
        Method of checking compatibility of two variants as nodes. Takes a Graph
        containing Edge data and an A and a B node to check.
    """
    if compatibility_kind is None and incompatibility_kind is None:
        raise ValueError("Compatibility and incompatibility kind can't both be None.")
    if compatibility_kind and incompatibility_kind:
        raise ValueError("Compatibility and incompatibility kind can't both be set.")

    def is_compatible(graph: Graph, var1: Node, var2: Node) -> bool:
        """Check whether these two nodes are compatible."""
        compatible = (
            True
            if compatibility_kind is None
            else any(e.kind == compatibility_kind for e in graph[var1.name, var2.name])
            or any(e.kind == compatibility_kind for e in graph[var2.name, var1.name])
        )
        if not compatible:
            return False
        incompatible = (
            False
            if incompatibility_kind is None
            else any(e.kind == incompatibility_kind for e in graph[var1.name, var2.name])
            or any(e.kind == incompatibility_kind for e in graph[var2.name, var1.name])
        )
        return not incompatible

    return is_compatible


def get_applicability_method(
    applicability_kind: Optional[str] = "applicability",
    inapplicability_kind: Optional[str] = None,
) -> Callable[[Graph, Node, Node], bool]:
    """Get a method dat signals whether an element is applicable when a certain
    constraint is active.

    Arguments:
        applicability_kind: Optional Edge kind that signals variant applicability.
        inapplicability_kind: Optional Edge kind that signals variant inapplicability.

    Returns:
        Method of checking applicability of an element for a constraint. Takes a Graph
        containing Edge data and an element and a constraint node to check.
    """
    if applicability_kind is None and inapplicability_kind is None:
        raise ValueError("Applicability and inapplicability kind can't both be None.")
    if applicability_kind and inapplicability_kind:
        raise ValueError("Applicability and inapplicability kind can't both be set.")

    def is_applicable(graph: Graph, element: Node, constraint: Node) -> bool:
        """Check whether these two nodes are compatible."""
        applicable = (
            True
            if applicability_kind is None
            else any(e.kind == applicability_kind for e in graph[element.name, constraint.name])
            or any(e.kind == applicability_kind for e in graph[constraint.name, element.name])
        )
        if not applicable:
            return False
        inapplicable = (
            False
            if inapplicability_kind is None
            else any(e.kind == inapplicability_kind for e in graph[element.name, constraint.name])
            or any(e.kind == inapplicability_kind for e in graph[constraint.name, element.name])
        )
        return not inapplicable

    return is_applicable


def get_score_method(
    variant_agg: str = "sum",
    config_agg: str = "sum",
    weights: Optional[List[str]] = None,
) -> Callable[[Tuple[Node, ...]], float]:
    """Get a configuration scoring method.

    Arguments:
        variant_agg: Variant node weights aggregation method. Either "sum" or "product".
        config_agg: Variant nodes' score aggregation method. Either "sum" or "product".
        weights: Optional selection of node weights to take into account.

    Returns:
        A method accepting a tuple of variant nodes and returning a single
        performance score.
    """

    def get_score(config: Tuple[Node, ...]) -> float:
        """Get the performance score of a configuration of element variants."""
        variant_scores = [get_variant_score(n, weights, variant_agg) for n in config]
        config_score = get_configuration_score(variant_scores, config_agg)
        return config_score

    return get_score


class CompatibilityAnalysis:
    """Compatibility analysis class.

    The input graph is expected to contain nodes that represent *variants* of certain
    (functional) elements. A variant is a variation of an element that fulfills the same
    functionality as other variants of that elements.

    Arguments:
        graph: Graph containing compatibility data between different variants.
        variants: Dictionary of element names to lists of variant nodes.
        interface_method: An optional method accepting a Graph and two variant nodes
            that returns whether the two given variants have an interface. Acts as a
            guard for whether compatibility checking is necessary when
            `interface_compatibility` is set.
        compatibility_method: A method accepting a Graph and two variant nodes to check
            whether the two given variants are compatible (by returning `True`).
        interface_compatibility: When set, the compatibility of two variants is only
            checked when they share an interface.
        score_method: A method accepting a tuple of variant nodes and
            returning a single performance score. See
            [`get_score_method`][ragraph.analysis.compatibility.get_score_method].
        constraints: Optional list of constraint nodes ("project scope") for which
            variants must be applicable in order to be taken into account.
        applicability_method: A method accepting a graph, a variant node and a
            constraint node that returns whether a variant is applicable when that
            constraint is set.
    """

    style = generic.Style(
        palettes=generic.Palettes(
            fields=dict(
                compatible=generic.FieldPalette(categorical="green"),
                incompatible=generic.FieldPalette(categorical="red"),
            )
        ),
        piemap=generic.PieMapStyle(
            display="labels",
            fields=["compatible", "incompatible"],
            scale_weight="scale",
            mode="relative",
        ),
        labels=generic.LabelsStyle(shorten=False),
    )

    def __init__(
        self,
        graph: Graph,
        variants: Dict[str, List[Node]],
        interface_method: Optional[Callable[[Graph, Node, Node], bool]] = None,
        compatibility_method: Callable[[Graph, Node, Node], bool] = get_compatibility_method(
            compatibility_kind="compatibility", incompatibility_kind=None
        ),
        interface_compatibility: bool = True,
        score_method: Callable[[Tuple[Node, ...]], float] = get_score_method(
            variant_agg="sum", config_agg="sum", weights=None
        ),
        constraints: Optional[List[Node]] = None,
        applicability_method: Callable[[Graph, Node, Node], bool] = get_applicability_method(
            applicability_kind="applicability", inapplicability_kind=None
        ),
    ):
        self.graph = graph
        """Graph containing compatibility data between different variants."""

        self.interface_method = interface_method
        """An optional method accepting a Graph and two variant nodes that returns whether the two
        given variants have an interface. Acts as a guard for whether compatibility checking is
        necessary when `interface_compatibility` is set."""

        self.compatibility_method = compatibility_method
        """A method accepting a Graph and two variant nodes to check whether the two given variants
        are compatible (by returning `True`)."""

        self.interface_compatibility = interface_compatibility
        """When set, the compatibility of two variants is only checked when they share an interface.
        """

        self.applicability_method = applicability_method
        """A method accepting a graph, a variant node and a constraint node that returns whether a
        variant is applicable when that constraint is set."""

        self.score_method = score_method
        """A method accepting a tuple of variant nodes and returning a single performance score.
        See [`get_score_method`][ragraph.analysis.compatibility.get_score_method]."""

        self._elements: List[str] = []
        self._element_nums: List[int] = []
        self._disabled_elements: List[str] = []

        self._constraints: List[Node] = [] if constraints is None else constraints

        self._variants: Dict[str, List[Node]] = dict()
        self._variants_list: List[Node] = []
        self.variants = variants

    @property
    def variants(self) -> Dict[str, List[Node]]:
        """Variants as lists by element name."""
        return self._variants

    @variants.setter
    def variants(self, value: Dict[str, List[Node]]):
        self._variants = value
        self._elements = []
        self._element_nums = []
        self._variants_list = []
        for el, variants in value.items():
            variants = [v for v in value[el] if self.is_applicable(v)]
            if variants:
                self._elements.append(el)
            else:
                self._disabled_elements.append(el)
                continue
            self._variants_list.extend(sorted(variants, key=lambda x: x.name))
            self._element_nums.append(len(variants))

    @property
    def constraints(self) -> List[Node]:
        """List of constraint nodes."""
        return self._constraints

    @constraints.setter
    def constraints(self, value: List[Node]):
        self._constraints = value
        self.variants = self._variants  # Re-trigger variants setter.

    @property
    def elements(self) -> List[str]:
        """Enabled elements with at least one applicable variant.
        See `self.constraints` and `self.applicability_method`.
        """
        return self._elements

    @property
    def disabled_elements(self) -> List[str]:
        """Disabled elements with no applicable variant.
        See `self.constraints` and `self.applicability_method`.
        """
        return self._disabled_elements

    @property
    def element_nums(self) -> List[int]:
        """Number of applicable variants per element."""
        return self._element_nums

    @property
    def variants_list(self) -> List[Node]:
        """Flat list of applicable variants sorted by (enabled) element."""
        return self._variants_list

    def is_applicable(self, variant: Node) -> bool:
        """Whether a variant is applicable to the currently set of constraints."""
        if self.constraints:
            return all(
                self.applicability_method(self.graph, variant, constraint)
                for constraint in self.constraints
            )
        return True

    def has_interface(self, var1: Node, var2: Node) -> bool:
        """Whether two variants have an interface that needs a compatibility check."""
        return (self.interface_method is None) or self.interface_method(self.graph, var1, var2)

    def is_compatible(self, var1: Node, var2: Node) -> bool:
        """Whether the two given variants are compatible according to
        [`self.applicability_method`][ragraph.analysis.compatibility.CompatibilityAnalysis.applicability_method]
        """

        if self.interface_compatibility:
            return (
                self.compatibility_method(self.graph, var1, var2)
                if self.has_interface(var1, var2)
                else True
            )

        return self.compatibility_method(self.graph, var1, var2)

    def get_compatibility_matrix(
        self, variants: Optional[List[Node]] = None
    ) -> Union["np.ndarray", List[List[float]]]:
        """Compatibility matrix between variants.

        Arguments:
            variants: Optional list of variants to return the compatibility matrix for.
                Also see `compatibility_method`.

        Returns:
            Compatibility matrix as a list of lists or numpy array (if numpy is
            available).
        """
        variants = self._variants_list if variants is None else variants
        dim = len(variants)
        mat = (
            np.zeros((dim, dim)) if HAVE_NUMPY else [[0.0 for j in range(dim)] for i in range(dim)]
        )
        for i in range(dim):
            var1 = variants[i]
            for j in range(i + 1, dim):
                if self.is_compatible(var1, variants[j]):
                    mat[i][j] = 1.0
                    mat[j][i] = 1.0
        return mat

    def yield_feasible_configurations(self) -> Generator[Tuple[Node, ...], None, None]:
        """Yield the feasible configurations for this problem."""
        variants = self._variants_list
        compat = self.get_compatibility_matrix(variants)
        configs = yield_feasible_configurations(compat, self._element_nums)
        for config in configs:
            yield tuple(variants[idx] for idx in config)

    def get_config_score(self, config: Tuple[Node, ...]) -> float:
        """Score a configuration. Does NOT check if the supplied config is valid!

        Arguments:
            config: Configuration as a tuple of variant nodes.

        Returns:
            Configuration score.
        """
        return self.score_method(config)

    def yield_scored_configurations(
        self,
    ) -> Generator[Tuple[Tuple[Node, ...], float], None, None]:
        """Yield the feasible configurations and their scores."""
        for config in self.yield_feasible_configurations():
            yield config, self.get_config_score(config)

    def get_ranked_configurations(
        self, descending: bool = True
    ) -> List[Tuple[Tuple[Node, ...], float]]:
        """Get the feasible configurations, sorted by score.

        Arguments:
            descending: Whether highest scores should go first.

        Returns:
            Sorted tuples of configurations and scores.
        """
        return sorted(self.yield_scored_configurations(), key=lambda x: x[-1], reverse=descending)

    def write_csv(self, path: Union[str, Path], scored: bool = True, limit: int = -1) -> int:
        """Write feasible configurations and optional scores to a CSV file."""
        path = Path(path)
        num = 0
        with open(path, mode="w+", encoding="utf-8") as f:
            f.write(";".join(self._elements))

        def append(content: str):
            with open(path, mode="a", encoding="utf-8") as f:
                f.write(content)

        if scored:
            gen = self.yield_scored_configurations()
            append(";score\n")
            if limit > 0:
                while num < limit:
                    cfg, score = next(gen)
                    append(";".join(n.name for n in cfg))
                    append(f";{score}\n")
                    num += 1
            else:
                for cfg, score in gen:
                    append(";".join(n.name for n in cfg))
                    append(f";{score}\n")
                    num += 1
        else:
            cfgs = self.yield_feasible_configurations()
            append("\n")
            if limit > 0:
                while num < limit:
                    cfg = next(cfgs)
                    append(";".join(n.name for n in cfg))
                    append("\n")
                    num += 1
            else:
                for cfg in cfgs:
                    append(";".join(n.name for n in cfg))
                    append("\n")
                    num += 1
        return num

    def get_plot_graph(self) -> Graph:
        """Get a plot ready graph for the current compatibility analysis problem."""
        g = Graph(add_parents=True, add_children=True)
        variants = deepcopy(self.variants_list)
        for variant in variants:
            g.add_node(variant)

        for i, el1 in enumerate(self.elements[:-1]):
            variants1 = self.variants[el1]
            for el2 in self.elements[i + 1 :]:
                variants2 = self.variants[el2]
                for var1 in variants1:
                    for var2 in variants2:
                        e1 = Edge(
                            g[var1.name],
                            g[var2.name],
                            kind="compatibility",
                            labels=[],
                        )
                        if self.has_interface(var1, var2):
                            e1.labels.append("interface")
                            e1.labels.append(
                                "compatible" if self.is_compatible(var1, var2) else "incompatible"
                            )
                        else:
                            e1.labels.append("no interface")
                            e1.labels.append("compatible")
                            e1.weights["scale"] = 0.25
                        e2 = Edge(
                            g[var2.name],
                            g[var1.name],
                            kind="compatibility",
                            labels=e1.labels.copy(),
                            weights=e1.weights.copy(),
                        )
                        g.add_edge(e1)
                        g.add_edge(e2)
        return g

    def plot(self, **mdm_args) -> "go.Figure":
        """Visualize the compatibility analysis problem."""
        g = self.get_plot_graph()
        variants = [g[n.name] for n in self.variants_list]

        # Merge styles if provided.
        if "style" in mdm_args:
            style = deepcopy(self.style)
            style.update(mdm_args["style"])
            mdm_args["style"] = style
        else:
            mdm_args["style"] = self.style

        return mdm(leafs=variants, edges=g.edges, **mdm_args)



"""Sequencing utils."""
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

import numpy as np

from ragraph.analysis import logger
from ragraph.analysis._classes import BranchsortAnalysis
from ragraph.analysis.sequence.metrics import net_markov_flow_adjacency
from ragraph.graph import Graph, Node

branchsort_analysis = BranchsortAnalysis(
    "Branchsort analysis",
    description="""Sequence all branches in a hierarchy using any sequencing algorithm. Returns
    the graph, (sequenced) root nodes and sequenced leaf nodes.

    Note:
        Setting just the `root` argument will sort it's children (recursively if you want). Setting
        `nodes` is like setting multiple roots. Setting `leafs` treats the given nodes as leaf
        nodes, whose children will not be reordered.

    Note:
        It is assumed that all edges are already nicely instantiated for all their parent
        components. E.g., your root node should have it's own incoming and outgoing edges.
    """,
)


@branchsort_analysis
def branchsort(
    graph: Graph,
    algo: Callable,
    algo_args: Optional[Dict[str, Any]] = None,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[Node], List[str]]] = None,
    leafs: Optional[Union[List[Node], List[str]]] = None,
    edge_weights: Optional[List[str]] = None,
    inherit: bool = True,
    loops: bool = False,
    inplace: bool = True,
    recurse: bool = True,
    names: bool = False,
    safe: bool = True,
) -> Tuple[Graph, List[Node], List[Node]]:
    """docstring stub"""
    # Sort roots
    assert isinstance(algo_args, dict)
    graph, node_sequence = algo(
        graph=graph,
        root=None,
        nodes=nodes,
        inherit=inherit,
        loops=loops,
        edge_weights=edge_weights,
        inplace=inplace,
        **algo_args,
    )
    assert node_sequence

    # Start with empty leaf sequence.
    leaf_sequence: List[Node] = []

    # Traverse branches
    for node in node_sequence:
        branchsort_analysis.log(f"Traversing '{node.name}'...")

        # Nothing to do for leaf nodes or nodes that should be treated as leafs.
        if node.is_leaf or node in leafs:  # type: ignore
            branchsort_analysis.log(f"Skipped leaf node '{node.name}'.")
            leaf_sequence.append(node)
            continue

        # Recursive case, get sorted children as the sorted "roots" of recursive call.
        if recurse:
            branchsort_analysis.log(f"Recursing into '{node.name}'...")
            _, _child_sequence, _leaf_sequence = branchsort(
                algo=algo,
                algo_args=algo_args,
                graph=graph,
                root=node,
                leafs=leafs,
                inherit=inherit,
                loops=loops,
                edge_weights=edge_weights,
                inplace=inplace,
                recurse=recurse,
            )
            leaf_sequence.extend(_leaf_sequence)
            branchsort_analysis.log(
                f"Leaf sequence extended with {[n.name for n in _leaf_sequence]}."
            )

        # Non-recursive case, get sorted children in this call.
        else:
            branchsort_analysis.log("Non-recursive call to algorithm...")
            _, _leaf_sequence = algo(
                graph=graph,
                root=node,
                inherit=inherit,
                loops=loops,
                edge_weights=edge_weights,
                inplace=inplace,
                **algo_args,
            )
            leaf_sequence.extend(_leaf_sequence)
            branchsort_analysis.log(
                f"Leaf sequence extended with {[n.name for n in _leaf_sequence]}."
            )

    return graph, node_sequence, leaf_sequence


def markov_decision(
    graph: Graph,
    options: List[Node],
    inherit: bool = True,
    loops: bool = False,
    inf: float = 1.0,
    dep: float = 1.0,
    mu: float = 1.5,
    context: Optional[List[Node]] = None,
) -> int:
    """Make a decision based on a Markov flow analysis of the adjacency matrix. The node
    with the lowest net markov flow is picked.

    Arguments:
        graph: Graph data.
        options: Nodes to decide between.
        inf: The weight to subtract outgoing flow by.
        dep: The weight to add incoming flow by.
        mu: Evaporation constant when calculating flow through nodes.
        context: Optional superset of nodes with respect to the options argument that
            constitutes the "complete" Markov chain that should be taken into account.

    Returns:
        Index of node to pick.
    """
    if context:
        adj = graph.get_adjacency_matrix(context, inherit=inherit, loops=loops)
        mapping = [context.index(option) for option in options]
        penalties = net_markov_flow_adjacency(adj, inf=inf, dep=dep, mu=mu)[mapping]
    else:
        adj = graph.get_adjacency_matrix(options, inherit=inherit, loops=loops)
        penalties = net_markov_flow_adjacency(adj, inf=inf, dep=dep, mu=mu)
    idx = np.argmin(penalties)
    logger.debug(
        f"Markov decision:\nNode {options[idx].name} has the lowest of penalties of "
        + f"{[n.name for n in options]}.\n{penalties}"
    )
    return np.argmin(penalties)



"""# RaGraph analysis utilities

Utility functions for managing the graph object during analysis.
"""

from copy import copy
from typing import Iterable, List, Optional, Union

from ragraph.graph import Graph
from ragraph.node import Node


def create_parent(
    graph: Graph,
    children: Union[List[Node], List[str]],
    prefix: str = "node",
    lower: int = 0,
    **node_kwargs,
) -> Node:
    """Add a parent node for given children to a given graph.

    Arguments:
        graph: Graph to create parent in.
        children: Children to add a parent node for.
        prefix: Parent name prefix, will be appended by first available integer.
        lower: Lower bound for integer to append to parent name prefix.
        kwargs:

    Returns:
        Created parent node.
    """
    if isinstance(children[0], str):
        parent_kind = graph.node_dict[children[0]].kind
    else:
        parent_kind = children[0].kind
    child_nodes = [
        graph.node_dict[child] if isinstance(child, str) else child for child in children
    ]
    if "name" not in node_kwargs:
        if parent_kind:
            prefix = ".".join([parent_kind, prefix])

        node_kwargs["name"] = get_available_name(graph, prefix, lower)
    parent = Node(kind=parent_kind, **node_kwargs)
    graph.add_node(parent)
    set_children(graph, parent, child_nodes)
    return parent


def get_available_name(graph: Graph, prefix: str, lower: int) -> str:
    """Get an available node name starting with prefix, appended the first integer
    that is free starting at lower.

    Arguments:
        graph: Graph to check availability in.
        prefix: Node name prefix, will be appended by first available integer.
        lower: Lower bound for integer to append to name prefix.

    Returns:
        First available node name not in [`graph.node_dict`][ragraph.graph.Graph.node_dict].
    """
    available = False
    i = lower
    while not available:
        name = prefix + str(i)
        available = name not in graph.node_dict
        i += 1

    return name


def inherit_edges(
    graph: Graph,
    new_parents: Iterable[Union[str, Node]],
    parent_siblings: Iterable[Union[str, Node]],
) -> None:
    """Let a cluster root nodes inherit all edges from their direct children
    (not descendants).

    Arguments:
        graph: Graph to work in.
        new_parents: New parent nodes to recreate edges for.
        parent_siblings: Already existing parent siblings.

    Note:
        Sometimes a clustering iteration does not introduce a new parent for every node,
        hence the distinction in parent nodes. New parent nodes look for edges from/to
        their children where old parent nodes already have edges (as this method has
        probably been called before). In general, it's better to wait with edge
        recreation until the graph's hierarchy is complete. Inserting and moving nodes
        half-way a tree would have way too many implications for the edges.
    """
    new_parent_nodes = [
        graph.node_dict[node] if isinstance(node, str) else node for node in new_parents
    ]
    parent_sibling_nodes = [
        graph.node_dict[node] if isinstance(node, str) else node for node in parent_siblings
    ]
    assert set(new_parent_nodes).isdisjoint(
        parent_sibling_nodes
    ), "There should be no overlap between new parents and siblings."

    for i in new_parent_nodes:
        for j in new_parent_nodes:
            if i == j:
                continue
            for child_edge_ij in graph.edges_between_all(i.children, j.children):
                edge_ij = copy(child_edge_ij)
                edge_ij.name = None  # type: ignore
                edge_ij.source = i
                edge_ij.target = j
                graph.add_edge(edge_ij)

    for i in new_parent_nodes:
        for j in parent_sibling_nodes:
            for child_edge_ij in graph.edges_between_all(i.children, [j]):
                edge_ij = copy(child_edge_ij)
                edge_ij.name = None  # type: ignore
                edge_ij.source = i
                edge_ij.target = j
                graph.add_edge(edge_ij)
            for child_edge_ji in graph.edges_between_all([j], i.children):
                edge_ji = copy(child_edge_ji)
                edge_ji.name = None  # type: ignore
                edge_ji.source = j
                edge_ji.target = i
                graph.add_edge(edge_ji)


def set_children(graph: Graph, parent: Node, children: List[Node]) -> None:
    """Safely set children of parent node.

    Arguments:
        graph: Graph to work in.
        parent: Parent node.
        children: Children to set.

    Note:
        If the children are attempted to be set to the parent itself, return.
        If children has length 1, the grandchildren are used and child is deleted.
        If children has length 1 and it is a leaf node, a ValueError is raised.
    """
    if not children:
        return

    if len(children) == 1:
        child = children[0]
        if parent == child:
            return
        if child.is_leaf:
            raise ValueError("Cannot set a single leaf node as children.")
        child.parent = parent
        graph.del_node(child, inherit=True)
        return

    for child in children:
        child.parent = parent


def unset_parent(graph: Graph, node: Node, roots: Optional[List[Node]] = None) -> None:
    """Clearing the parent of a node. Moves sole siblings upwards, replacing the parent.

    Arguments:
        graph: Graph to work in.
        node: The node of which the parent must be cleared.
        roots: Roots are protected nodes that are never deleted.

    Note:
        The node is has its parent unset and is therefore removed from the parent's
        children. The parent may have just a single other child left, in which case
        that child node should replace the parent altogether. When it has more children,
        it is kept intact.

    Example:
        Resetting the parent of "a" in the following tree:

        ```
               g
           ____|____
          |         |
          e         f
         _|_     ___|___
        |   |   |   |   |
        a   b   c   d   e
        ```

        leaves the following forest:

        ```
        a        g
             ____|____
            |         |
            b         f
                   ___|___
                  |   |   |
                  c   d   e
        ```

        in which "a" has become a root and "b" has become a child of "g" since "e" has
        been removed.
    """
    parent = node.parent
    if not parent:
        return
    roots = roots or []

    # Remove this node's parent and reset bus prop.
    node.parent = None
    node.is_bus = False

    # Move any potential sole children upwards if possible.
    if parent not in roots and len(parent.children) == 1:
        sole = parent.children[0]
        sole.is_bus = False
        grandparent = parent.parent
        if grandparent:
            # Move up to grandparent's children
            grandparent.children = grandparent.children + parent.children
        else:
            # Sole remaining child becomes a root as well.
            parent.children[0].parent = None
        graph.del_node(parent, inherit=False)


def clear_local_hierarchy(
    graph: Graph, leafs: List[Node], roots: Optional[List[Node]] = None
) -> None:
    """Strip a local hierarchy between a given root node and leaf nodes, removing all
    non-leaf nodes and setting the root as parent.

    Arguments:
        graph: Graph to strip nodes in.
        leafs: Lower bound leaf nodes to start from.
        root: Optional root nodes that are kept intact.
    """
    roots = roots or []
    for leaf in leafs:
        leaf.is_bus = False
        while leaf.parent:
            if leaf.parent in roots:
                leaf.parent = None
            else:
                graph.del_node(leaf.parent, inherit=True)



"""# GRIP format support"""
from pathlib import Path
from typing import Union
from uuid import UUID

from lxml import etree

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node


def from_grip(path: Union[str, Path]) -> Graph:
    """Decode GRIP XML file, string, or element into a Graph.

    Arguments:
        path: GRIP XML file path.

    Returns:
        Graph object.
    """

    tree = etree.parse(str(path))
    root = tree.getroot()

    graph = Graph(
        uuid=root.attrib.get("WorkspaceID"),
        name=root.attrib.get("WorkspaceName"),
        annotations=dict(root.attrib),
    )

    parse_collection(graph, root, "Objecten", "Object", "object")
    parse_objectenboom(graph, root)

    parse_collection(graph, root, "Functies", "Functie", "functie")
    parse_collection(graph, root, "Systeemeisen", "Systeemeis", "systeemeis")
    parse_collection(graph, root, "Scope", "Scope", "scope")
    parse_collection(graph, root, "Raakvlakken", "Raakvlak", "raakvlak")

    parse_systeemeis_edges(graph, root)
    parse_object_edges(graph, root)
    parse_scope_edges(graph, root)
    parse_raakvlak_edges(graph, root)

    return graph


def parse_collection(graph: Graph, root: etree.Element, collection: str, item: str, kind: str):
    coll = root.find(collection)
    for el in coll.iterfind(item):
        annotations = dict(el.attrib)
        if el.find("ID") is not None:
            id1 = el.find("ID").attrib.get("ID1")
            name = el.attrib.get("Name")
            if name:
                name = f"{name} | {id1}"
            else:
                name = id1
        else:
            name = el.attrib.get("Name")
        graph.add_node(
            Node(
                uuid=el.attrib.get("GUID"),
                name=name,
                kind=kind,
                annotations=annotations,
            )
        )


def parse_objectenboom(graph: Graph, root: etree.Element):
    collection = root.find("Objecten")
    for el in collection.iterfind("Object"):
        parent = graph[UUID(el.attrib.get("GUID"))]
        for sub in el.iterfind("SI_Onderliggend"):
            for obj in sub.iterfind("ObjectOnderliggend"):
                child_id = UUID(obj.attrib.get("GUID"))
                graph[child_id].parent = parent


def parse_systeemeis_edges(graph: Graph, root: etree.Element, kind="systeemeis"):
    elems = root.find("Systeemeisen").iterfind("Systeemeis")
    for el in elems:
        source = graph[UUID(el.attrib.get("GUID"))]

        for me_eis in el.iterfind("CI_MEEisObject"):
            eis_obj = me_eis.find("EisObject")
            eis_obj.attrib.get("GUID")
            object_id = eis_obj.find("SI_Object").find("Object").attrib.get("GUID")
            target = graph[UUID(object_id)]
            graph.add_edge(Edge(source, target, kind=kind))
            graph.add_edge(Edge(target, source, kind=kind))


def parse_object_edges(graph: Graph, root: etree.Element, kind="object"):
    collection = root.find("Objecten")
    for el in collection.iterfind("Object"):
        source = graph[UUID(el.attrib.get("GUID"))]

        for sub in el.iterfind("SI_Functie"):
            for functie in sub.iterfind("Functie"):
                functie_id = UUID(functie.attrib.get("GUID"))
                target = graph[functie_id]
                graph.add_edge(Edge(source, target, kind=kind))
                graph.add_edge(Edge(target, source, kind=kind))


def parse_scope_edges(graph: Graph, root: etree.Element, kind="scope"):
    elems = root.find("Scope").iterfind("Scope")
    for el in elems:
        source = graph[UUID(el.attrib.get("GUID"))]

        for eis in el.iterfind("SI_Systeemeis"):
            eis_id = eis.find("Systeemeis").attrib.get("GUID")
            target = graph[UUID(eis_id)]
            graph.add_edge(Edge(source, target, kind=kind))
            graph.add_edge(Edge(target, source, kind=kind))

        for functie in el.iterfind("SI_Functie"):
            functie_id = functie.find("Functie").attrib.get("GUID")
            target = graph[UUID(functie_id)]
            graph.add_edge(Edge(source, target, kind=kind))
            graph.add_edge(Edge(target, source, kind=kind))

        for raakvlak in el.iterfind("SI_Raakvlak"):
            raakvlak_id = raakvlak.find("Raakvlak").attrib.get("GUID")
            target = graph[UUID(raakvlak_id)]
            graph.add_edge(Edge(source, target, kind=kind))
            graph.add_edge(Edge(target, source, kind=kind))

        for obj in el.iterfind("SI_Object"):
            obj_id = obj.find("Object").attrib.get("GUID")
            target = graph[UUID(obj_id)]
            graph.add_edge(Edge(source, target, kind=kind))
            graph.add_edge(Edge(target, source, kind=kind))


def parse_raakvlak_edges(graph: Graph, root: etree.Element, kind="raakvlak"):
    elems = root.find("Raakvlakken").iterfind("Raakvlak")
    for el in elems:
        raakvlak = graph[UUID(el.attrib.get("GUID"))]

        objecten = [
            graph[UUID(item.find("Objecttype").attrib.get("GUID"))]
            for item in el.iterfind("SI_Objecttype")
        ]

        functies = [
            graph[UUID(item.find("Functie").attrib.get("GUID"))]
            for item in el.iterfind("SI_Functie")
        ]

        for i, obj in enumerate(objecten):
            graph.add_edge(Edge(raakvlak, obj, kind=kind))
            graph.add_edge(Edge(obj, raakvlak, kind=kind))
            for func in functies:
                graph.add_edge(Edge(obj, func, kind=kind))
                graph.add_edge(Edge(func, obj, kind=kind))

            for other in objecten[i + 1 :]:
                graph.add_edge(Edge(obj, other, kind=kind))
                graph.add_edge(Edge(other, obj, kind=kind))

        for func in functies:
            graph.add_edge(Edge(raakvlak, func, kind=kind))
            graph.add_edge(Edge(func, raakvlak, kind=kind))



"""# Edge class module"""

from typing import Any, Dict, List, Optional, Union
from uuid import UUID

from ragraph.generic import Annotations, Metadata
from ragraph.node import Node


class Edge(Metadata):
    """Edge between a source [`Node`][ragraph.node.Node] and a target [`Node`][ragraph.node.Node].

    Arguments:
        source: Source [`Node`][ragraph.node.Node] of this edge.
        target: Target [`Node`][ragraph.node.Node] of this edge.
        name: Instance name. Given a UUID if none provided.
        kind: Kind or main category of this instance.
        labels: Labels categorizing this instance.
        weights: Dictionary of weights attached to this instance.
        annotations: Miscellaneous properties of this instance.
        uuid: Fixed UUID if desired, generated when left set to None.
    """

    def __init__(
        self,
        source: Node,
        target: Node,
        name: Optional[str] = None,
        kind: str = "edge",
        labels: Optional[List[str]] = None,
        weights: Optional[Dict[str, Union[int, float]]] = None,
        annotations: Union[Annotations, Dict[str, Any], None] = None,
        uuid: Optional[Union[str, UUID]] = None,
    ):
        Metadata.__init__(
            self,
            name=name,
            kind=kind,
            labels=labels,
            weights=weights,
            annotations=annotations,
            uuid=uuid,
        )
        self._source: Node
        self._target: Node
        setattr(self, "source", source)
        setattr(self, "target", target)

    def __str__(self):
        return "{}(source='{}', target='{}', name='{}')".format(
            self.__class__.__name__, self.source.name, self.target.name, self.name
        )

    def __repr__(self):
        kwargdict = dict(
            source=str(self.source),
            target=str(self.target),
            name=repr(self.name),
            kind=repr(self.kind),
            labels=self.labels,
            weights=self.weights,
            annotations=self.annotations,
            uuid=repr(self.uuid),
        )
        kwargs = ", ".join([str(k) + "=" + str(v) for k, v in kwargdict.items()])
        return "<{}.{}({}) at {}>".format(
            self.__class__.__module__, self.__class__.__name__, kwargs, hex(id(self))
        )

    @property
    def source(self) -> Node:
        """Edge source node."""
        return self._source

    @source.setter
    def source(self, value: Node):
        self._source = value

    @property
    def target(self) -> Node:
        """Edge target node."""
        return self._target

    @target.setter
    def target(self, value: Node):
        self._target = value

    @property
    def json_dict(self) -> Dict[str, Any]:
        """JSON dictionary representation.

        Returns:
            source: Source node UUID (not Node) as str.
            target: Target node UUID (not Node) as str.
            kind: Kind as str.
            labels: Labels as list of str.
            weights: Weights as dict.
            annotations: Annotations as a dictionary.
        """
        return self.as_dict(use_uuid=True)

    def as_dict(self, use_uuid: bool = False) -> Dict[str, Any]:
        """Return a copy as a (serializable) dictionary.

        Arguments:
            use_uuid: Whether to use UUIDs instead of names.

        Returns:
            source: Source node name or UUID (not Node) as str.
            target: Target node name or UUID (not Node) as str.
            kind: Kind as str.
            labels: Labels as list of str.
            weights: Weights as dict.
            annotations: Annotations as a dictionary.
            uuid: UUID as str if toggled.
        """
        if use_uuid:
            return dict(
                source=str(self.source.uuid),
                target=str(self.target.uuid),
                name=self.name,
                kind=self.kind,
                labels=self.labels,
                weights=self.weights,
                annotations=self.annotations.as_dict(),
                uuid=str(self.uuid),
            )
        else:
            return dict(
                source=self.source.name,
                target=self.target.name,
                name=self.name,
                kind=self.kind,
                labels=self.labels,
                weights=self.weights,
                annotations=self.annotations.as_dict(),
            )



"""# CSV format support"""

import csv
import warnings
from collections import OrderedDict
from pathlib import Path
from typing import Any, Dict, List, Optional, Union

from ragraph.generic import Annotations
from ragraph.graph import Edge, Graph, Node

NODE_COLUMNS = ["name", "uuid", "kind", "labels", "parent", "children", "is_bus"]
EDGE_COLUMNS = ["source", "target", "name", "uuid", "kind", "labels"]


def from_csv(
    nodes_path: Optional[Union[str, Path]] = None,
    edges_path: Optional[Union[str, Path]] = None,
    csv_delimiter: str = ";",
    iter_delimiter: str = ";",
    node_weights: Optional[List[str]] = None,
    edge_weights: Optional[List[str]] = None,
    **graph_kwargs: Any,
) -> Graph:
    """Convert CSV files to a graph.

    Nodes file (optional) requires at least a `name` column. Optional special
    property columns are `kind`, `labels`, `parent`, `children` and `is_bus`.

    Edges file requires at least a `source` and `target` column. Optional special
    property columns are `kind` and `labels`.

    Automatically completes one sided parent-child relationships.

    Arguments:
        nodes_path: Path of optional nodes CSV file.
        edges_path: Path of edges CSV file.
        csv_delimiter: Delimiter of CSV file.
        iter_delimiter: Iterable delimiter (i.e. for children names list).
        node_weights: Columns to interpret as node weights.
        edge_weights: Columns to interpret as edge weights.
        **graph_kwargs: Optional [`Graph`][ragraph.graph.Graph] arguments when instantiating.

    Returns:
        Graph object.

    Raises:
        InconsistencyError: if graph is inconsistent.
    """
    graph = Graph(**graph_kwargs)

    if not nodes_path and not edges_path:
        return graph

    if not node_weights:
        node_weights = []

    if not edge_weights:
        edge_weights = []

    if nodes_path:
        node_dict = _load_nodes(nodes_path, csv_delimiter, iter_delimiter, node_weights)
    else:
        node_dict = _derive_nodes(edges_path, csv_delimiter)

    if edges_path:
        edges = _load_edges(edges_path, csv_delimiter, iter_delimiter, node_dict, edge_weights)
    else:
        edges = []

    graph.add_parents = True
    graph.add_children = True
    graph.nodes = node_dict
    graph.edges = edges
    graph.check_consistency(raise_error=True)
    return graph


def to_csv(
    graph: Graph,
    stem_path: str,
    csv_delimiter: str = ";",
    iter_delimiter: str = ";",
    use_uuid: bool = False,
):
    """Save graph to nodes and edges CSV files.

    Arguments:
        graph: Graph to save.
        stem_path: Stem path for output CSV's. Appended with _nodes.csv and _edges.csv.
        csv_delimiter: CSV delimiter.
        iter_delimiter: Iterable delimiter (i.e. for children names list).
        use_uuid: Whether to export UUIDs, too.
    """
    stem_path = Path(stem_path)

    nodes_fname = stem_path.name + "_nodes.csv"
    nodes_path = stem_path.with_name(nodes_fname)
    _save_nodes(graph, nodes_path, csv_delimiter, iter_delimiter, use_uuid=use_uuid)

    edges_fname = stem_path.name + "_edges.csv"
    edges_path = stem_path.with_name(edges_fname)
    _save_edges(graph, edges_path, csv_delimiter, iter_delimiter, use_uuid=use_uuid)


def _load_nodes(
    fpath: Union[str, Path],
    csv_delimiter: str,
    iter_delimiter: str,
    node_weights: List[str],
) -> Dict[str, Node]:
    """Convert node CSV file to a list of nodes.

    Arguments:
        fpath: Path to nodes CSV file.
        csv_delimiter: CSV delimiter.
        iter_delimiter: Iterable delimiter (i.e. for children names list).
        node_weights: Columns to interpret as node weights.

    Returns:
        Node dictionary (name, node).
    """
    with Path(fpath).open(newline="") as f:
        reader = csv.DictReader(f, delimiter=csv_delimiter)
        columns = reader.fieldnames
        rows = [row for row in reader]
    if "name" not in columns:
        raise ValueError(
            "Nodes CSV file {} needs at least a 'name' column, found {}.".format(
                str(fpath), columns
            )
        )

    wght_cols = node_weights
    anno_cols = [col for col in columns if col not in NODE_COLUMNS and col not in wght_cols]

    # Handle node names and annotations
    names = []
    nodes = []
    for row in rows:
        name = row["name"]
        anno = Annotations(**{col: row[col] for col in anno_cols})
        weights = {col: _convert_to_num(row[col]) for col in wght_cols}
        names.append(name)
        uuid = row.get("uuid", None)
        nodes.append(Node(name, weights=weights, annotations=anno, uuid=uuid))

    if len(names) != len(set(names)):
        raise ValueError("Node names are not unique.")

    # Update other values including references
    node_dict = OrderedDict(zip(names, nodes))
    for row in rows:
        name = row["name"]
        node = node_dict[name]

        kind = row.get("kind", None)
        if kind:
            node.kind = str(kind)

        labels = row.get("labels", None)
        if labels:
            node.labels = [s.strip() for s in labels.split(iter_delimiter)]

        parent = row.get("parent", None)
        if parent:
            node.parent = node_dict[parent]

        children = row.get("children", [])
        if children:
            children_names = [n.strip() for n in children.split(iter_delimiter)]
            children = [node_dict[n] for n in children_names if node_dict[n] not in node.children]
            node.children = node.children + children

        is_bus = row.get("is_bus", "false")
        if is_bus.strip().lower() == "true":
            node.is_bus = True

    return node_dict


def _derive_nodes(edges_path: Union[str, Path], csv_delimiter: str) -> Dict[str, Node]:
    """Derive nodes from edges CSV file.

    Arguments:
        edges_path: Path to edges CSV file.
        csv_delimiter: CSV delimiter.

    Returns:
        Node dictionary (name, node).
    """

    with Path(edges_path).open(newline="") as f:
        reader = csv.DictReader(f, delimiter=csv_delimiter)
        columns = reader.fieldnames
        rows = [row for row in reader]
    if "source" not in columns or "target" not in columns:
        raise ValueError(
            "Edges CSV file {} needs at least a 'source' and 'target' column.".format(
                str(edges_path)
            )
        )

    nodes = []
    names = []
    for row in rows:
        source = row["source"]
        if source not in names:
            nodes.append(Node(source))
            names.append(source)
        target = row["target"]
        if target not in names:
            nodes.append(Node(target))
            names.append(target)

    return OrderedDict(zip(names, nodes))


def _load_edges(
    fpath: Union[str, Path],
    csv_delimiter: str,
    iter_delimiter: str,
    node_dict: List[Node],
    edge_weights: List[str],
) -> List[Edge]:
    """Convert edge CSV file to graph.

    Arguments:
        fpath: Path to edges CSV file.
        csv_delimiter: CSV delimiter.
        iter_delimiter: Iterable delimiter (i.e. for children names list).
        node_dict: Node dictionary (name, node).
        edge_weights: Columns to interpret as edge weights.

    Returns:
        Edge list.
    """

    with Path(fpath).open(newline="") as f:
        reader = csv.DictReader(f, delimiter=csv_delimiter)
        columns = reader.fieldnames
        rows = [row for row in reader]
    if "source" not in columns or "target" not in columns:
        raise ValueError(
            "Edges CSV file {} needs at least a 'source' and 'target' column.".format(str(fpath))
        )

    wght_cols = edge_weights
    anno_cols = [col for col in columns if col not in EDGE_COLUMNS and col not in wght_cols]

    edges = []
    for row in rows:
        if row.get("labels", None):
            row["labels"] = [s.strip() for s in row["labels"].split(iter_delimiter)]
        weights = {col: _convert_to_num(row[col]) for col in wght_cols}
        anno = Annotations(**{col: row[col] for col in anno_cols})
        kwargs = {col: row[col] for col in columns if col in EDGE_COLUMNS}
        kwargs["source"] = node_dict[kwargs["source"]]
        kwargs["target"] = node_dict[kwargs["target"]]
        edges.append(Edge(**kwargs, weights=weights, annotations=anno))

    return edges


def _save_nodes(
    graph: Graph,
    fpath: Path,
    csv_delimiter: str,
    iter_delimiter: str,
    use_uuid: bool = False,
):
    """Save graph nodes to CSV file.

    Arguments:
        graph: Graph to save.
        fpath: Path to nodes CSV file.
        csv_delimiter: CSV delimiter.
        iter_delimiter: Iterable delimiter (i.e. for children names list).
        use_uuid: Whether to export UUIDs, too.
    """
    # Weights column mapping
    wght_keys = [key for key in graph.node_weight_labels if key is not None]
    wght_map = {key: str(key) for key in wght_keys}
    wght_map.update({key: str(key) + "_weight" for key in wght_keys if str(key) in NODE_COLUMNS})

    # Annotations column mapping
    anno_keys = sorted(set([key for node in graph.node_list for key in node.annotations.keys()]))
    anno_map = {key: str(key) for key in anno_keys}
    anno_map.update(
        {
            key: str(key) + "_annotation"
            for key in anno_keys
            if str(key) in NODE_COLUMNS or str(key) in wght_map.values()
        }
    )

    row_dict = dict()
    for key in NODE_COLUMNS:
        row_dict[key] = None
    for key in wght_map.values():
        row_dict[key] = None
    for key in anno_map.values():
        row_dict[key] = None
    if not use_uuid:
        row_dict.pop("uuid")

    fpath = Path(fpath)
    with fpath.open(mode="w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=row_dict.keys(), delimiter=csv_delimiter)
        writer.writeheader()

        for node in graph.node_list:
            row_dict["name"] = node.name
            if use_uuid:
                row_dict["uuid"] = str(node.uuid)
            row_dict["kind"] = node.kind
            row_dict["labels"] = iter_delimiter.join(node.labels)
            if node.parent:
                row_dict["parent"] = node.parent.name
            else:
                row_dict["parent"] = None
            child_str = iter_delimiter.join([c.name for c in node.children])
            if child_str:
                row_dict["children"] = child_str
            else:
                row_dict["children"] = None
            row_dict["is_bus"] = node.is_bus

            for key, value in wght_map.items():
                row_dict[value] = node.weights[key]

            for key, value in anno_map.items():
                row_dict[value] = getattr(node.annotations, key, None)

            writer.writerow(row_dict)


def _save_edges(
    graph: Graph,
    fpath: Path,
    csv_delimiter: str,
    iter_delimiter: str,
    use_uuid: bool = False,
):
    """Save graph edges to CSV file.

    Arguments:
        graph: Graph to save.
        fpath: Path to edges CSV file.
        csv_delimiter: CSV delimiter.
        iter_delimiter: Iterable delimiter (i.e. for children names list).
        use_uuid: Whether to export UUIDs, too.
    """
    # Weights column mapping
    wght_keys = [key for key in graph.edge_weight_labels if key is not None]
    wght_map = {key: str(key) for key in wght_keys}
    wght_map.update({key: str(key) + "_weight" for key in wght_keys if str(key) in EDGE_COLUMNS})

    # Annotations column mapping
    anno_keys = sorted(set([key for edge in graph.edge_list for key in edge.annotations.keys()]))
    anno_map = {key: str(key) for key in anno_keys}
    anno_map.update(
        {
            key: str(key) + "_annotation"
            for key in anno_keys
            if str(key) in NODE_COLUMNS or str(key) in wght_map.values()
        }
    )

    row_dict = dict()
    for key in EDGE_COLUMNS:
        row_dict[key] = None
    for key in wght_map.values():
        row_dict[key] = None
    for key in anno_map.values():
        row_dict[key] = None
    if not use_uuid:
        row_dict.pop("uuid")

    fpath = Path(fpath)
    with fpath.open(mode="w", newline="") as f:
        writer = csv.DictWriter(f, fieldnames=row_dict.keys(), delimiter=csv_delimiter)
        writer.writeheader()

        for edge in graph.edge_list:
            row_dict["name"] = edge.name
            row_dict["source"] = edge.source.name
            row_dict["target"] = edge.target.name
            row_dict["kind"] = edge.kind
            row_dict["labels"] = iter_delimiter.join(edge.labels)
            if use_uuid:
                row_dict["uuid"] = str(edge.uuid)

            for key, value in wght_map.items():
                row_dict[value] = edge.weights[key]

            for key, value in anno_map.items():
                row_dict[value] = getattr(edge.annotations, key, None)

            writer.writerow(row_dict)


def _convert_to_num(value: str) -> Union[float, bool]:
    """Convert a string value to a float number or bool.

    Argument:
        value: The value to be converted.
    """
    if not isinstance(value, str):
        return value

    if value.strip().lower() == "true":
        return True
    elif value.strip().lower() == "false":
        return False
    else:
        try:
            # String is a number
            num = float(value)
            return num
        except ValueError:
            msg = " ".join(
                [
                    "You are assigning a string as a weight property.",
                    "Please use a float, integer, or a Bool.",
                    "A default value of 1.0 is used as a replacement for {}.".format(value),
                ]
            )
            warnings.warn(msg)
            return 1.0



"""Sequence metrics."""

from typing import List, Optional, Tuple, Union

import numpy as np

from ragraph.graph import Graph
from ragraph.node import Node


def feedback_marks(mat: np.ndarray) -> Tuple[float, np.ndarray]:
    """Measure the sum of feedback marks in an adjacency matrix.

    Arguments:
        mat: Adjacency matrix to analyze.

    Returns:
        Resulting metric.
        Cell contribution matrix.

    References:
        Steward, D. V., 1981, Systems Analysis and Management: Structure, Strategy and
            Design, Petrocelli Books, New York.
        Kusiak, A., & Wang, J. (1993). Decomposition of the Design Process. Journal of
            Mechanical Design, 115(4), 687. https://doi.org/10.1115/1.2919255
    """
    contrib = np.triu(mat, 1)
    return contrib.sum(), contrib


def feedback_distance(mat: np.ndarray) -> Tuple[float, np.ndarray]:
    """Measure the feedback length, e.g. distance from the adjacency matrix diagonal.

    Arguments:
        mat: Adjacency matrix to analyze.

    Returns:
        Resulting metric.
        Cell contribution matrix.

    References:
        Gebala, D. A., & Eppinger, S. D. (1991). Methods for analyzing design
            procedures. ASME Design Technical Conferences (Design Theory and
            Methodology), 227-233. Miami.
    """
    dim = len(mat)

    dimtile = np.tile(np.arange(dim), (dim, 1))
    ddist = dimtile - dimtile.T

    contrib = np.triu(mat, 1) * ddist
    return contrib.sum(), contrib


def lower_left(mat: np.ndarray) -> Tuple[float, np.ndarray]:
    """Measure the distance to the lower left of the adjacency matrix.

    Arguments:
        mat: Adjacency matrix to analyze.

    Returns:
        Resulting metric.
        Cell contribution matrix.

    References:
        Todd, D. S. (1997). Multiple criteria genetic algorithms in engineering design
            and operation. University of Newcastle.
    """
    dim = len(mat)

    dimtile = np.tile(np.arange(dim), (dim, 1))
    lldist = dimtile - dimtile.T + dim - 1

    contrib = mat * lldist
    return contrib.sum(), contrib


def feedback_lower_left(
    mat: np.ndarray, fb: float = 100.0, ff: float = 1.0
) -> Tuple[float, np.ndarray]:
    """Jointly measure lower left distance to the adjacency matrix diagonal and
    feedback marks. Feedback and feedforward are penalized differently, but both via a
    quadratic lower-left distance factor.

    Arguments:
        mat: Adjacency matrix to analyze.
        fb: Feedback adjacency multiplier.
        ff: Feedforward adjacency multiplier.

    Returns:
        Resulting metric.
        Cell contribution matrix.

    Note:
        Feedback marks above the diagonal are weighted 100 times more than those below
        the diagonal. The multiplier is offset by (n-1)^2.

    References:
        Scott, J. A. (1999). A strategy for modelling the design-development phase of a
            product. University of Newcastle. See Equation 6.1 and Figure 6.4.
    """
    dim = len(mat)

    dimtile = np.tile(np.arange(1, dim + 1), (dim, 1))
    lldist = dimtile - dimtile.T + dim
    omega = np.square(lldist)

    # Lower triangle
    omgl = ff * omega
    tril = np.tril(mat, -1)
    resl = omgl * tril

    # Upper triangle
    omgu = fb * omega
    triu = np.triu(mat, 1)
    resu = omgu * triu

    contrib = resl + resu
    return contrib.sum(), contrib


def feedback_crossover(
    mat: np.ndarray, fb: float = 0.9, co: float = 0.1
) -> Tuple[float, Tuple[np.ndarray, np.ndarray]]:
    """Jointly measure feedback marks and crossovers in an adjacency matrix.

    Arguments:
        mat: Adjacency matrix to analyze.
        fb: Weight of feedback marks.
        co: Weight of feedback crossovers.

    Returns:
        Resulting metric.
        Cell contribution matrices (feedback, crossover).

    Note:
        Crossovers are where the "axis" from feedback marks towards the diagonal cross,
        except when this crossing is a feedback mark in itself. Multiple lines from
        "above" in the matrix are summed, those from the right are **NOT** (treated as
        binary).

    References:
        McCulley, C., and Bloebaum, C. L., 1996, “A Genetic Tool for Optimal Design
            Sequencing in Complex Engineering Systems,” Struct. Optim., 12(2), pp.
            186-201.
    """
    # Treat binary matrix
    mat = (mat > 0) * 1.0

    # contrib_fb is upper triangularized
    total_fb, contrib_fb = feedback_marks(mat)

    # vertical cum sum (number of accumulated marks from the top until diagonal)
    vcumsum = np.triu(np.cumsum(contrib_fb, axis=0), 1)

    # horizontal "line" multiplier, 1s toward the left from the rightmost 1 except the
    # marks themselves:
    # rows of [0,0,1,0,0,1,0,0,0]
    # become  [1,1,0,1,1,0,0,0,0]
    hline = np.cumsum(contrib_fb[:, ::-1], axis=1)[:, ::-1]
    hline = (hline > 0) * 1.0 - contrib_fb
    # no need to np.triu() here, since vcumsum already is upper triangularized.

    # Contribution matrix at crossover points and total score.
    contrib_co = vcumsum * hline
    total_co = contrib_co.sum()

    return fb * total_fb + co * total_co, (contrib_fb, contrib_co)


def net_markov_flow_adjacency(
    mat: np.ndarray,
    inf: float = 1.0,
    dep: float = 1.0,
    mu: float = 2.0,
    scale: bool = True,
) -> np.ndarray:
    """Calculate a flow balance as if the matrix would constitute a (weighted)
    Markov chain.

    Arguments:
        mat: Adjacency matrix to analyze.
        inf: The weight to subtract outgoing flow by.
        dep: The weight to add incoming flow by.
        mu: Evaporation constant when calculating flow through nodes.
        scale: Whether to scale the inflow vector according to the adjacency matrix
            column sums.

    Returns:
        Nodal flow balance as an array.
    """
    from ragraph.analysis.cluster._markov import MarkovFlow

    flow = MarkovFlow(mat, mu, scale).flow_matrix[:-1, :-1]  # Exclude sink.
    inf_vector = flow.sum(axis=0)
    dep_vector = flow.sum(axis=1)
    return dep * dep_vector - inf * inf_vector


def net_markov_flow_graph(
    graph: Graph,
    nodes: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = True,
    loops: bool = False,
    inf: float = 1.0,
    dep: float = 1.0,
    mu: float = 2.0,
) -> np.ndarray:
    """Calculate a flow balance as if the graph would constitute a (weighted)
    Markov chain.

    Arguments:
        graph: Graph to analyze nodes of.
        nodes: Set of node (names) to calculate the flow with.
        inherit: Whether to take into account edges between descendants of nodes.
        loops: Whether to take into account self-loop edges.
        inf: The weight to subtract outgoing flow by.
        dep: The weight to add incoming flow by.
        mu: Evaporation constant when calculating flow through nodes.

    Returns:
        Nodal flow balance as an array.

    Note:
        Uses edge inheritance to calculate an adjacency matrix.
    """
    mat = graph.get_adjacency_matrix(nodes, inherit=inherit, loops=loops)
    return net_markov_flow_adjacency(mat, inf=inf, dep=dep, mu=mu)



"""Tarjan's Depth First Search Algorithm."""
from typing import Dict, List, Optional, Tuple, Union

from ragraph.analysis._classes import SequenceAnalysis
from ragraph.graph import Graph
from ragraph.node import Node

tarjans_dfs_analysis = SequenceAnalysis(
    "Tarjan's Depth First Search Algorithm",
    description="""\
    Sort an directed acyclic graph (DAG) in it's sequential order.

    Note:
        Can only parse nodes which form a directed acyclic graph (DAG).
    """,
)


@tarjans_dfs_analysis
def tarjans_dfs(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[str], List[Node]]] = None,
    inherit: bool = True,
    loops: bool = False,
    edge_weights: Optional[List[str]] = None,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, List[Node]]:
    """docstring stub"""
    assert nodes
    out = []
    unmarked = [n for n in nodes if isinstance(n, Node)]
    state = {n: "todo" for n in nodes}

    targets_of: Dict[Node, List[Node]] = {
        source: [
            target
            for target in unmarked
            if any(graph.edges_between(source, target, inherit=inherit, loops=False))
        ]
        for source in unmarked
    }

    def visit(n: Node):
        s = state[n]
        if s == "done":
            return
        if s == "temp":
            raise ValueError("Not a Directed Acyclic Graph (DAG).")
        state[n] = "temp"
        for m in targets_of[n]:
            visit(m)
        state[n] = "done"
        out.append(n)

    while unmarked:
        node = unmarked.pop()
        if state[node] == "done":
            continue
        visit(node)

    return graph, out[::-1]



"""#Similarity analysis utilities"""
from typing import Any, Callable, List


def on_hasattrs(attrs: List[str]) -> Callable[[Any], List[bool]]:
    """Get an object description function that checks whether an instance possesses certain
    attributes. It does not check the values thereof!

    Arguments:
        attrs: List of attributes to check the existence of.

    Returns:
        Object description function indicating attribute possession.
    """
    return lambda obj: [hasattr(obj, attr) for attr in attrs]


def on_checks(checks: List[Callable[[Any], bool]]) -> Callable[[Any], List[bool]]:
    """Get an object description function that runs a predefined set of checks (which should be in a
    fixed order) and returns their boolean results.

    Arguments:
        checks: Checks to perform.

    Returns:
        Object description function indicating check passings.
    """
    return lambda obj: [check(obj) for check in checks]


def on_hasweights(weights: List[str], threshold: float = 0.0) -> Callable[[Any], List[bool]]:
    """Check whether an objects has certain weights above a threshold in its weights dictionary
    property.

    Arguments:
        weights: Keys to the `obj.weights` dictionary to check.
        threshold: Threshold to verify against.

    Returns:
        Object description function indicating weights exceeding a threshold.
    """
    return lambda obj: [obj.weights.get(w, 0.0) >= threshold for w in weights]


def on_contains(contents: List[Any]) -> Callable[[Any], List[bool]]:
    """Check whether an object contains certain contents.

    Arguments:
        contents: Contents to check for with `lambda x: x in obj`.

    Returns:
        Object description function indicating content presence.
    """
    return lambda obj: [content in obj for content in contents]



"""Sorting algorithm based on the node dependency/influence in a Markov chain.
"""
from textwrap import dedent
from typing import List, Optional, Tuple, Union

import numpy as np

from ragraph.analysis._classes import Bound, Parameter, SequenceAnalysis
from ragraph.analysis.sequence.metrics import net_markov_flow_adjacency
from ragraph.graph import Graph, Node

markov_params = {
    p.name: p
    for p in [
        Parameter(
            "inf",
            float,
            description="Weight of relative node influence when sorting.",
            default=1.0,
            lower=Bound(0.0, inclusive=True, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        ),
        Parameter(
            "dep",
            float,
            description="Weight of relative node dependency when sorting.",
            default=1.0,
            lower=Bound(0.0, inclusive=True, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        ),
        Parameter(
            "mu",
            float,
            "Decay coefficient (usually 1.5 - 3.5). Influence or dependency of nodes "
            "decays according to this coefficient at each node it passes.",
            default=1.5,
            lower=Bound(1.0, inclusive=False, report="error"),
            upper=Bound(10.0, inclusive=True, report="warn"),
        ),
        Parameter(
            "scale",
            bool,
            "Whether to scale in-flow with respect to adjacency column sums.",
            default=False,
        ),
    ]
}

markov_sequencing_analysis = SequenceAnalysis(
    "Markov sequencing",
    description=dedent(
        """\
    Sequencing based on relative influence and dependency as if the graph were a
    Markov chain. Internally, this converts the edge weights between nodes into a flow
    (steady state probability) distribution matrix.
    """
    ),
    parameters=markov_params,
)


@markov_sequencing_analysis
def markov(
    graph: Graph,
    root: Optional[Union[str, Node]] = None,
    nodes: Optional[Union[List[Node], List[str]]] = None,
    edge_weights: Optional[List[str]] = None,
    inf: float = markov_params["inf"].default,  # type: ignore
    dep: float = markov_params["dep"].default,  # type: ignore
    mu: float = markov_params["mu"].default,  # type: ignore
    scale: bool = markov_params["scale"].default,  # type: ignore
    inherit: bool = True,
    loops: bool = False,
    inplace: bool = True,
    names: bool = False,
    safe: bool = True,
    **kwargs,
) -> Tuple[Graph, List[Node]]:
    """docstring stub"""
    assert nodes is not None
    adj = graph.get_adjacency_matrix(nodes=nodes, inherit=inherit, loops=loops, only=edge_weights)
    assert isinstance(adj, np.ndarray)

    # Obtain node penalty scores.
    penalties = net_markov_flow_adjacency(adj, inf=inf, dep=dep, mu=mu, scale=scale)

    # Calculate final sequence.
    idxs = np.argsort(penalties)

    # Reorder nodes
    seq = [nodes[i] for i in idxs]

    return graph, seq



"""RaGraph plot utilities."""
from typing import Dict, Iterable, List, Optional, Union

from plotly import graph_objects as go
from plotly.subplots import make_subplots

from ragraph.analysis.sequence._axis import get_axis_sequence  # noqa
from ragraph.edge import Edge
from ragraph.node import Node
from ragraph.plot.components.blank import Blank
from ragraph.plot.components.labels import Labels
from ragraph.plot.components.legend import Legend
from ragraph.plot.components.piemap import PieMap
from ragraph.plot.components.tree import Tree
from ragraph.plot.generic import Component, Style


def get_subplots(components: List[List[Optional[Component]]], style: Style = Style()) -> go.Figure:
    """Get a subplots [`plotly.graph_objects.Figure`][plotly.graph_objects.Figure] for the given
    components list of lists.

    Arguments:
        components: Components to be laid out as subplots based on their width and height
            properties.
        style: Style options.
    """
    rows = len(components)

    components_t = list(zip(*components))  # Transpose helper matrix.
    cols = len(components_t)

    min_x_ranges = [
        min(
            [comp.xaxis.range[0] for comp in col if comp and comp.xaxis.range],
            default=0,
        )
        if any(col)
        else 0
        for col in components_t
        if components_t
    ]

    max_x_ranges = [
        max(
            [comp.xaxis.range[1] for comp in col if comp and comp.xaxis.range],
            default=1,
        )
        if any(col)
        else 1
        for col in components_t
    ]

    widths = [(x_max - x_min) * style.boxsize for (x_max, x_min) in zip(max_x_ranges, min_x_ranges)]

    min_y_ranges = [
        min(
            [comp.yaxis.range[0] for comp in row if comp and comp.yaxis.range],
            default=0,
        )
        if any(row)
        else 0
        for row in components
    ]

    max_y_ranges = [
        max(
            [comp.yaxis.range[1] for comp in row if comp and comp.yaxis.range],
            default=1,
        )
        if any(row)
        else 1
        for row in components
    ]

    heights = [
        (y_max - y_min) * style.boxsize for (y_max, y_min) in zip(max_y_ranges, min_y_ranges)
    ]

    fig = make_subplots(
        rows=rows,
        cols=cols,
        shared_xaxes=True,
        shared_yaxes=True,
        horizontal_spacing=0,
        vertical_spacing=0,
        column_widths=None if sum(widths) == 0 else widths,
        row_heights=None if sum(heights) == 0 else heights,
    )
    fig.layout.update(style.layout)

    shapes, annotations = [], []
    for i, row in enumerate(components):
        y = "y" if i == 0 else f"y{i*cols+1}"
        for j, component in enumerate(row):
            if not component:
                component = Blank()

            x = "x" if i + j == 0 else f"x{j+1+i*cols}"

            # Add traces.
            for trace in component.traces:
                fig.add_trace(trace, i + 1, j + 1)

            # Set shapes' reference axis
            for shape in component.shapes:
                shape.update({"xref": x, "yref": y})

            # Add annotations.
            for annotation in component.annotations:
                annotation.update({"xref": x, "yref": y})

            shapes.extend(component.shapes)
            annotations.extend(component.annotations)

            component.xaxis.update(range=(min_x_ranges[j], max_x_ranges[j]))
            component.yaxis.update(range=(min_y_ranges[i], max_y_ranges[i]))

            component.width = widths[j]
            component.height = heights[i]

            # Axis overrides
            fig.update_xaxes(row=i + 1, col=j + 1, patch=component.xaxis)
            fig.update_yaxes(row=i + 1, col=j + 1, patch=component.yaxis)

    fig.layout.shapes = shapes
    fig.layout.annotations = annotations

    margin = style.layout.margin
    fig.layout.update(
        {
            "width": sum(widths) + margin["l"] + margin["r"],
            "height": sum(heights) + margin["t"] + margin["b"],
        }
    )

    return fig


def process_fig(fig: go.Figure, style: Style = Style(), show: bool = True) -> Optional[go.Figure]:
    """Show figure with config if `show` is set, otherwise return figure unchanged.

    Arguments:
        fig: Plotly figure.
        style: Style containing additional config.
        show: Whether to show the figure inline.
    """
    if show:
        style.config["toImageButtonOptions"] = dict(
            format="svg",
            filename="ragraph_plot",
            width=fig.layout.width,
            height=fig.layout.height,
            margin=dict(l=0, t=0, r=0, b=0),
        )
        fig.show(config=style.config)
        return None
    else:
        return fig


def get_mdm_grid(
    leafs: List[Node], edges: List[Edge], style: Style = Style()
) -> List[List[Optional[Component]]]:
    """Get grid layout for mdm figure.

    Arguments
        leafs: List of nodes to be displayed.
        edges: The edges to be displayed.
        style: Plot style option mapping.

    Returns
        Grid of go.Figure objects.
    """
    col_number_row: List[Optional[Component]] = []
    piemap_row: List[Optional[Component]] = []
    grid: List[List[Optional[Component]]] = []

    if style.row_col_numbers:
        style.labels.textorientation = "vertical"
        col_number_row = [
            None,
            None,
            None,
            Labels([Node(str(i + 1)) for i in range(len(leafs))], style=style),
        ]

        style.labels.textorientation = "horizontal"
        piemap_row = [
            Tree(leafs, style=style),
            Labels(leafs, style=style),
            Labels([Node(str(i + 1)) for i in range(len(leafs))], style=style),
            PieMap(rows=leafs, cols=leafs, edges=edges, style=style),
        ]
    else:
        piemap_row = [
            Tree(leafs, style=style),
            Labels(leafs, style=style),
            PieMap(rows=leafs, cols=leafs, edges=edges, style=style),
        ]

    if style.show_legend and edges:
        if col_number_row:
            col_number_row.append(None)
        piemap_row.append(Legend(edges, style=style))

    if col_number_row:
        grid.append(col_number_row)
    grid.append(piemap_row)

    return grid


def get_dmm_grid(
    rows: List[Node], cols: List[Node], edges: List[Edge], style: Style = Style()
) -> List[List[Union[go.Figure, None]]]:
    """Get grid layout for mdm figure.

    Arguments
        rows: The nodes to be placed on the rows of the matrix.
        cols: The columns to be placed on the columns of the matrix.
        edges: The edges to be displayed.
        style: Plot style option mapping.

    Returns
        Grid of go.Figure objects.
    """
    grid: List[List[Optional[Component]]] = []
    col_num_row: List[Optional[Component]] = []
    col_label_row: List[Optional[Component]] = []
    piemap_row: List[Optional[Component]] = []
    if style.row_col_numbers:
        style.labels.textorientation = "vertical"
        col_label_row = [
            None,
            None,
            Labels(cols, style=style),
        ]
        col_num_row = [
            None,
            None,
            Labels([Node(str(i + 1)) for i in range(len(cols))], style=style),
        ]

        style.labels.textorientation = "horizontal"
        piemap_row = [
            Labels(rows, style=style),
            Labels([Node(str(i + 1)) for i in range(len(rows))], style=style),
            PieMap(rows=rows, cols=cols, edges=edges, style=style),
        ]
    else:
        style.labels.textorientation = "vertical"
        col_label_row = [None, Labels(cols, style=style)]

        style.labels.textorientation = "horizontal"
        piemap_row = [
            Labels(rows, style=style),
            PieMap(rows=rows, cols=cols, edges=edges, style=style),
        ]

    if style.show_legend and edges:
        col_label_row.append(None)
        if col_num_row:
            col_num_row.append(None)
        piemap_row.append(Legend(edges, style=style))

    grid.append(col_label_row)
    if col_num_row:
        grid.append(col_num_row)
    grid.append(piemap_row)

    return grid


def get_swatchplot(*args: Iterable[List[str]], **kwargs: Dict[str, List[str]]) -> go.Figure:
    """Swatch plot of colormaps.

    Arguments:
        *args: Hex coded color lists.
        **kwargs: Names to hex coded color lists.

    Returns:
        Plotly figure.
    """
    colormaps = kwargs
    for i, colormap in enumerate(args):
        colormaps[str(i)] = colormap
    bars = [
        go.Bar(
            orientation="h",
            y=[name] * len(colors),
            x=[1] * len(colors),
            customdata=list(range(len(colors))),
            marker=dict(color=colors),
            hovertemplate="%{y}[%{customdata}] = %{marker.color}<extra></extra>",
        )
        for name, colors in colormaps.items()
    ]

    fig = go.Figure(
        data=bars[::-1],
        layout=dict(
            barmode="stack",
            barnorm="fraction",
            bargap=0.5,
            showlegend=False,
            xaxis=dict(range=[-0.02, 1.02], showticklabels=False, showgrid=False),
            height=max(600, 40 * len(colormaps)),
            margin=dict(b=10),
        ),
    )

    return fig



"""# Adjacency and mapping matrices support"""

from typing import Any, Dict, List, Optional, Union

from ragraph.edge import Edge
from ragraph.graph import Graph
from ragraph.node import Node

try:
    import numpy as np

except ImportError:
    np = None


def from_matrix(
    matrix: Union["np.ndarray", List[List[int]], List[List[float]]],
    rows: Optional[Union[List[Node], List[str]]] = None,
    cols: Optional[Union[List[Node], List[str]]] = None,
    weight_label: str = "default",
    empty: Optional[Union[int, float]] = 0.0,
    **graph_args: Dict[str, Any],
) -> Graph:
    """Create a graph from an adjacency or mapping matrix.

    Arguments:
        matrix: Matrix to convert into a graph.
        rows: Nodes or node labels corresponding to the rows of the matrix.
        cols: Nodes or node labels corresponding to the columns of the matrix. If none
            are provided, the row labels are re-used.
        weight_label: Weight label to use for matrix values.
        empty: Cell value to be considered "empty", e.g. no edge should be created.
        **graph_args: Additional arguments to [`Graph`][ragraph.graph.Graph] constructor.

    Returns:
        Graph object.

    Note:
        If no row labels are provided, they are generated in a "node#" format.
        If no column labels are provided, they are assumed to be equal to the rows.
        For non-square matrices, you should provide node labels!
    """

    rows = [Node(f"node{i}") for i in range(len(matrix))] if rows is None else rows

    # init empty graph
    graph = Graph(**graph_args)

    # Parse row labels and create new nodes.
    node_rows = _parse_labels(graph, rows)

    # Init nodes from provided labels.
    if cols is None:
        node_cols = node_rows
    else:
        node_cols = _parse_labels(graph, cols)

    # Dimension check of matrix and labels.
    dim = (len(matrix), len(matrix[0]))
    labdim = (len(node_rows), len(node_cols))
    if dim != labdim:
        raise ValueError(f"Matrix dimensions {dim} do not agree with label dimensions {labdim}.")

    # Generate edges and return.
    edges = [
        Edge(
            source,
            target,
            name=f"{source.name}->{target.name}",
            weights={weight_label: matrix[row][col]},
        )
        for row, target in enumerate(node_rows)
        for col, source in enumerate(node_cols)
        if matrix[row][col] != empty
    ]
    graph.edges = edges
    return graph


def to_matrix(
    graph: Graph,
    rows: Optional[Union[List[Node], List[str]]] = None,
    cols: Optional[Union[List[Node], List[str]]] = None,
    inherit: bool = False,
    loops: bool = False,
    only: Optional[List[str]] = None,
) -> Union[np.ndarray, List[List[float]]]:
    """Convert graph data into a directed numerical adjacency or mapping matrix.

    Arguments:
        graph: Graph to fetch data from.
        rows: Nodes representing the matrix rows.
        cols: Nodes representing the matrix columns if different from the rows.
        inherit: Whether to count weights between children of the given nodes.
        loops: Whether to calculate self-loops from a node to itself.
        only: Optional subset of edge weights to consider. See
            [`ragraph.edge.Edge`][ragraph.edge.Edge] for default edge weight implementation.

    Returns:
        Adjacency matrix as a 2D numpy array if numpy is present. Otherwise it will return a 2D
        nested list.

    Note:
        Note that the matrix is directed! Columns are inputs to rows.
    """
    if rows is None:
        rows = graph.leafs
    else:
        rows = [n if isinstance(n, Node) else graph.node_dict[n] for n in rows]

    if cols is None:
        cols = rows
    else:
        cols = [n if isinstance(n, Node) else graph.node_dict[n] for n in cols]

    dim = (len(rows), len(cols))
    if np:
        matrix = np.zeros(dim, dtype=float)
    else:
        matrix = [[0.0 for j in range(dim[1])] for i in range(dim[0])]

    for col, source in enumerate(cols):
        for row, target in enumerate(rows):
            if source == target and not loops:
                continue

            sources = [source]
            targets = [target]
            if inherit:
                sources.extend(source.descendants)
                targets.extend(target.descendants)

            weight = sum(
                [_get_weight(e, only=only) for e in graph.edges_between_all(sources, targets)]
            )
            matrix[row][col] = float(weight)

    return matrix


def _parse_labels(graph: Graph, labels: Union[List[Node], List[str]]) -> List[Node]:
    """Parse matrix labels into a list of nodes."""
    nodes: List[Node] = [
        label if isinstance(label, Node) else graph.node_dict.get(str(label), Node(str(label)))
        for label in labels
    ]

    for node in nodes:
        graph.add_node(node)  # Skips existing identical nodes.

    return nodes


def _get_weight(e: Edge, only: Optional[List[str]] = None) -> float:
    """Get numerical weight from an edge."""
    if only:
        return sum(e.weights.get(k, 0.0) for k in only)
    else:
        return e.weight
